
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dirent.h>

#ifdef	RHP
#include "rhp.h"
#else
#include "rhp_t.h"
#endif

#include "rif_f.h"

#define EPS	1e-9	/* for time interval */
#define LEN	2048

static void	Error(char *dir,char *name,char *msg)
{
	if (*dir!='\0') (void)fprintf(stderr,"%s/",dir);

	(void)fprintf(stderr,"%s : %s\n",name,msg); exit(1);
}

static int	StrCmp(char *s1,char *s2)
{
	int	c;

	while ((c=(*s1>='A' && *s1<='Z')?*s1|32:*s1)==(*s2)) {
	    if (c=='\0') return 0;

	    ++s1; ++s2;
	}
	return 1;
}

static void	Open(char *dir,char *name,int *Nx,int *Ny)
{
	char	path[LEN];

	(void)sprintf(path,"%s/%s",dir,name);
	(void)ReadImageFile_Float(path,Nx,Ny,NULL,NULL);
}

static void	Read(char *dir,char *name,int Nx,int Ny,FOM ***fom)
{
	char	path[LEN];
	int	x,y;

	(void)sprintf(path,"%s/%s",dir,name);
	(void)ReadImageFile_Float(path,&x,&y,fom,NULL);

	if (Nx!=x || Ny!=y) Error(dir,name,"image size not match.");
}

static int	Compare(OutputLog *OL1,OutputLog *OL2)
{
	int	it=OL2->it-OL1->it;

	return (it)?it:(OL1->it==0)?(OL1->c>OL2->c)?1:(OL1->c<OL2->c)?-1:0:
		       (OL1->it==1)?(OL1->a>OL2->a)?1:(OL1->a<OL2->a)?-1:0:0;
}

void	InitReadHiPic(char *dir,HiPic *hp)
{
	char		*env,q_tif[LEN],str[LEN],
			output_log[LEN]="",
			dark_tif[LEN]="",
			*fnf="file not found.",
			*nmfdd="no memory for directory data.",
			*bsn="bad sequence number.",
			*dsn="duplicated sequence number.";
	DIR		*Dir;
	struct dirent	*sd;
	int		l,q,it,x,y,i;
	FILE		*file;
	double		c,a,c1,c2;
	OutputLog	*OL;

	if ((env=getenv("RHP_O"))!=NULL) (void)strcpy(output_log,env);
	if ((env=getenv("RHP_D"))!=NULL) (void)strcpy(dark_tif,env);

	(void)sprintf(q_tif,
		      "%c[0-9]*[0-9].tif",
		      ((env=getenv("RHP_Q"))!=NULL && *env!='\0' &&
		       (*env|32)>='a' && (*env|32)<='z')?*env|32:'q');

	hp->Nq=0;

	if ((Dir=opendir(dir))==NULL) Error("",dir,"directory not open.");

#define NAME	sd->d_name

	while ((sd=readdir(Dir))!=NULL)
	    if (*output_log=='\0' && !StrCmp(NAME,"output.log"))
		(void)sprintf(output_log,"%s/%s",dir,NAME);
	    else if (*dark_tif=='\0' && !StrCmp(NAME,"dark.tif"))
		(void)strcpy(dark_tif,NAME);
	    else if ((NAME[0]|32)==q_tif[0] &&
		     (l=(int)strlen(NAME))>5 &&
		     !StrCmp(NAME+l-4,".tif") &&
		     sscanf(NAME+1,"%d",&q)==1 &&
		     q>=hp->Nq)
		hp->Nq=q+1;

	if (*output_log=='\0') Error(dir,"output.log",fnf);
	if (*dark_tif=='\0') Error(dir,"dark.tif",fnf);
	if (hp->Nq==0) Error(dir,q_tif,fnf);

#define MALLOC(type,noe)	(type *)malloc(sizeof(type)*(noe))

	if ((hp->dir=strdup(dir))==NULL ||
	    (hp->q_tif=MALLOC(char *,hp->Nq))==NULL) Error("",dir,nmfdd);

	rewinddir(Dir);

	for (q=0; q<hp->Nq; q++) hp->q_tif[q]=NULL;

	while ((sd=readdir(Dir))!=NULL)
	    if ((NAME[0]|32)==q_tif[0] &&
		(l=(int)strlen(NAME))>5 &&
		!StrCmp(NAME+l-4,".tif") &&
		sscanf(NAME+1,"%d",&q)==1 &&
		q>=0) {
		if (q>=hp->Nq) Error(dir,NAME,bsn);

		if (hp->q_tif[q]!=NULL) Error(dir,NAME,dsn);

		if ((hp->q_tif[q]=strdup(NAME))==NULL) Error(dir,NAME,nmfdd);
	    }

	(void)closedir(Dir);

	if ((hp->OL=MALLOC(OutputLog,hp->Nq+1))==NULL)
	    Error("",output_log,"no memory for log data.");

	for (q=0; q<=hp->Nq; q++) hp->OL[q].it=(-1);

	Open(dir,dark_tif,&(hp->Nx),&(hp->Ny));

	hp->Ni=hp->Nt=0;

	if (*output_log=='-')
	    file=stdin;
	else
	    if ((file=fopen(output_log,"r"))==NULL)
		Error("",output_log,"file not open.");

	while (fgets(str,LEN,file)!=NULL)
	    if (sscanf(str,"%d %lf %lf %d%n",&q,&c,&a,&it,&l)==4) {
		str[l]='\0';

		if (q<0 || q>=hp->Nq || hp->q_tif[q]==NULL) Error("",str,bsn);

		if (it!=0 && it!=1) Error("",str,"unacceptable log data.");

		Open(dir,hp->q_tif[q],&x,&y);

		if (hp->Nx!=x || hp->Ny!=y)
		    Error(dir,hp->q_tif[q],"image size not match.");

		if (hp->OL[q].it!=(-1)) Error("",str,dsn);

		if (it==0) ++(hp->Ni); else ++(hp->Nt);

		if (hp->Ni+hp->Nt==1) c1=c2=c; else if (c1>c) c1=c;
					       else if (c2<c) c2=c;

		hp->OL[q].q=q; hp->OL[q].c=c; hp->OL[q].a=a; hp->OL[q].it=it;
	    }
	    else
#ifndef	ONLY_CT_VIEWS
		if (file==stdin)
#endif
		break;

	if (file!=stdin) (void)fclose(file);

	if (hp->Ni==0) Error(dir,q_tif,"no file referred as I0-image.");
	if (hp->Nt==0) Error(dir,q_tif,"no file referred as I-image.");

	if ((hp->I=MALLOC(FOM **,hp->Ni+(hp->Ni==1)))==NULL ||
	    (hp->T=MALLOC(FOM *,hp->Ny))==NULL ||
	    (*(hp->T)=MALLOC(FOM,hp->Ny*hp->Nx))==NULL)
	    Error("",dir,"no memory for image data.");

	Read(dir,dark_tif,hp->Nx,hp->Ny,&(hp->D));

	qsort(hp->OL,hp->Nq+1,sizeof(OutputLog),(int (*)())Compare);

	OL=hp->OL+hp->Nt;
	for (i=0; i<hp->Ni; i++) {
	    if (i>0 && OL[i-1].c+EPS>=OL[i].c)
		Error("",output_log,"bad time interval.");

	    Read(dir,hp->q_tif[OL[i].q],hp->Nx,hp->Ny,&(hp->I[i]));
	}
	if (hp->Ni==1) {
	    hp->Ni=2; OL[1].q=OL[0].q;
		      OL[0].c=c1-EPS; OL[1].c=c2+EPS;
		      OL[1].a=OL[0].a; OL[1].it=OL[0].it; hp->I[1]=hp->I[0];
	}
	for (y=1; y<hp->Ny; y++) hp->T[y]=hp->T[y-1]+hp->Nx;
}

static void	Free(int Ny,FOM **fom)
{
	while (--Ny>=0) free(fom[Ny]); free(fom);
}

void	ReadHiPic(HiPic *hp,int t)
{
	char		str[LEN];
	OutputLog	*OLt,*OLi;
	FOM		**T,**I1,**I2;
	int		i,y,x;
	double		r1,r2,I_D,T_D;

	if (t<0 || t>=hp->Nt) {
	    (void)sprintf(str,"%d",t); Error("",str,"bad sequence number.");
	}
	OLt=hp->OL+t;
	Read(hp->dir,hp->q_tif[OLt->q],hp->Nx,hp->Ny,&T);

	OLi=hp->OL+hp->Nt;
	for (i=hp->Ni-2; i>0; i--) if (OLt->c>=OLi[i].c) break;

	I1=hp->I[i]; I2=hp->I[i+1];
	r1=1.0-(r2=(OLt->c-OLi[i].c)/(OLi[i+1].c-OLi[i].c));

	for (y=0; y<hp->Ny; y++)
	for (x=0; x<hp->Nx; x++)
	    hp->T[y][x]=((I_D=r1*I1[y][x]+r2*I2[y][x]-hp->D[y][x])>0.0 &&
			 (T_D=T[y][x]-hp->D[y][x])>0.0)?T_D/I_D:ERROR_VALUE;

	Free(hp->Ny,T);
}

void	TermReadHiPic(HiPic *hp)
{
	int	i,q,Ny=hp->Ny;

	if (hp->Ni==2) {
	    if (hp->I[1]!=hp->I[0]) Free(Ny,hp->I[1]); Free(Ny,hp->I[0]);
	}
	else
	    for (i=hp->Ni-1; i>=0; i--) Free(Ny,hp->I[i]);

	Free(Ny,hp->D); free(*(hp->T)); free(hp->T); free(hp->I); free(hp->OL);

	for (q=hp->Nq-1; q>=0; q--) if (hp->q_tif[q]!=NULL) free(hp->q_tif[q]);

	free(hp->q_tif); free(hp->dir);
}
