
#include <stdio.h>
#include <stdlib.h>
#include "rif_f.h"
#include "csi_f.h"
#include "sif_f.h"

extern void	Error(char *msg);

#define LEN	2048

static void	Read(char *path,
		     int *Nx,int *Ny,FOM ***fom,double *base,double *step)
{
	int	bps,n;
	char	*id,str[8][LEN],*u_c,
		*u="using integer PV as is floating-point PV",
		*c="converting PV by TIFF image description tag";
	double	f1,f2;

	if ((bps=ReadImageFile_Float(path,Nx,Ny,fom,&id))!=0) {
	    if (id!=NULL) {
		if ((n=sscanf(id,"%s %s %s %s %s %s %s %s",
				 str[0],str[1],str[2],str[3],
				 str[4],str[5],str[6],str[7]))==2 || n==6)
		    if (sscanf(str[n-2],"%lf",&f1)!=1 ||
			sscanf(str[n-1],"%lf",&f2)!=1 || f1>=f2)
			u_c=u;
		    else {
			u_c=c; *base=f1; *step=(f2-f1)/(double)((1ULL<<bps)-1);
		    }
		else
		    u_c=((n!=3 && n!=4 && n!=8) ||
			 sscanf(str[n-2],"%lf",base)!=1 ||
			 sscanf(str[n-1],"%lf",step)!=1 || *step<=0.0)?u:c;

		free(id);
	    }
	    else
		u_c=u;
#ifdef	U_C
	    (void)fprintf(stderr,"%s : %s (warning).\n",path,u_c);
#endif
	}
	else {
	    u_c=NULL; if (id!=NULL) free(id);
	}
	if (u_c!=c) {
	    *base=0.0; *step=1.0;
	}
}

static void	Span(int O,int N1,int N2,int *O1,int *O2,int *N)
{
	if (O<0) {
	    *O1=0; *O2=(-O); if ((*N=N2+O)>N1) *N=N1;
	}
	else {
	    *O1=O; *O2=0; if ((*N=N1-O)>N2) *N=N2;
	}
}

static void	Sub_2D_Float(char **argv)
{
	int	Nx1,Ny1,Nx2,Ny2,x,y,x1,x2,Nx,y1,y2,Ny;
	FOM	**F1,**F2;
	double	b1,s1,b2,s2,b;

	Read(argv[1],&Nx1,&Ny1,&F1,&b1,&s1);
	Read(argv[2],&Nx2,&Ny2,&F2,&b2,&s2); b=b1-b2;

	if ((x=atoi(argv[3]))<=(-Nx2) || x>=Nx1 ||
	    (y=atoi(argv[4]))<=(-Ny2) || y>=Ny1) Error("bad offset.");

	Span(x,Nx1,Nx2,&x1,&x2,&Nx);
	Span(y,Ny1,Ny2,&y1,&y2,&Ny);

	for (y=0; y<Ny; y++)
	for (x=0; x<Nx; x++) F1[y][x]=b+s1*F1[y+y1][x+x1]-s2*F2[y+y2][x+x2];

	StoreImageFile_Float(argv[5],Nx,Ny,F1,SIF_F_desc);
	(void)puts(SIF_F_desc);

	for (y=0; y<Ny2; y++) free(F2[y]); free(F2);
	for (y=0; y<Ny1; y++) free(F1[y]); free(F1);
}

static void	Sub_3D_Float(char **argv)
{
	char	**path1,**path2,form[LEN],path[LEN];
	int	Nx1,Ny1,Nz1,Nx2,Ny2,Nz2,x,y,z,
		X1,X2,Nx,Y1,Y2,Ny,z1,z2,Nz,y1,y2,x1,x2;
	FOM	**F,**F1,**F2;
	double	b1,s1,b2,s2,b;

	path1=CheckSliceImages_Float(argv[1],argv[2],&Nx1,&Ny1,&Nz1,NULL);
	path2=CheckSliceImages_Float(argv[3],argv[4],&Nx2,&Ny2,&Nz2,NULL);

	if ((x=atoi(argv[5]))<=(-Nx2) || x>=Nx1 ||
	    (y=atoi(argv[6]))<=(-Ny2) || y>=Ny1 ||
	    (z=atoi(argv[7]))<=(-Nz2) || z>=Nz1) Error("bad offset.");

	Span(x,Nx1,Nx2,&X1,&X2,&Nx);
	Span(y,Ny1,Ny2,&Y1,&Y2,&Ny);
	Span(z,Nz1,Nz2,&z1,&z2,&Nz);

	if ((F=(FOM **)malloc(sizeof(FOM *)*Ny))==NULL ||
	    (*F=(FOM *)malloc(sizeof(FOM)*Ny*Nx))==NULL)
	    Error("memory allocation error.");

	for (y=1; y<Ny; y++) F[y]=F[y-1]+Nx;

	x=1; for (z=10; z<Nz; z*=10) ++x;
	(void)sprintf(form,"%s/%%0%dd.tif",argv[8],x);

	for (z=0; z<Nz; z++) {
	    Read(path1[z+z1],&Nx1,&Ny1,&F1,&b1,&s1);
	    Read(path2[z+z2],&Nx2,&Ny2,&F2,&b2,&s2); b=b1-b2;

	    for (y=0, y1=Y1, y2=Y2;
		 y<Ny && y1<Ny1 && y2<Ny2;
		 y++, y1++, y2++) {
		for (x=0, x1=X1, x2=X2;
		     x<Nx && x1<Nx1 && x2<Nx2;
		     x++, x1++, x2++) F[y][x]=b+s1*F1[y1][x1]-s2*F2[y2][x2];

		if (x2==Nx2)
		    for (; x<Nx && x1<Nx1; x++, x1++) F[y][x]=b+s1*F1[y1][x1];
		else
		    for (; x<Nx && x2<Nx2; x++, x2++) F[y][x]=b-s2*F2[y2][x2];

		for (; x<Nx; x++) F[y][x]=b;
	    }
	    if (y2==Ny2)
		for (; y<Ny && y1<Ny1; y++, y1++) {
		    for (x=0, x1=X1;
			 x<Nx && x1<Nx1;
			 x++, x1++) F[y][x]=b+s1*F1[y1][x1];

		    for (; x<Nx; x++) F[y][x]=b;
		}
	    else
		for (; y<Ny && y2<Ny2; y++, y2++) {
		    for (x=0, x2=X2;
			 x<Nx && x2<Nx2;
			 x++, x2++) F[y][x]=b-s2*F2[y2][x2];

		    for (; x<Nx; x++) F[y][x]=b;
		}

	    for (; y<Ny; y++) for (x=0; x<Nx; x++) F[y][x]=b;

	    (void)sprintf(path,form,z);
            StoreImageFile_Float(path,Nx,Ny,F,SIF_F_desc);
            (void)printf("%d\t%s\n",z,SIF_F_desc);

	    for (y=0; y<Ny2; y++) free(F2[y]); free(F2);
	    for (y=0; y<Ny1; y++) free(F1[y]); free(F1);
	}
	free(*F); free(F);

	for (z=0; z<Nz2; z++) free(path2[z]); free(path2);
	for (z=0; z<Nz1; z++) free(path1[z]); free(path1);
}

int	main(int argc,char **argv)
{
	if (argc==6) Sub_2D_Float(argv); else
	if (argc==9) Sub_3D_Float(argv); else

	Error("usage : sub_float TIFF1 TIFF2 Ox Oy TIFF\n"
	      "        sub_float dir1/ NF1 dir2/ NF2 Ox Oy Oz dir/");

	return 0;
}
