
#include <stdio.h>
#include <stdlib.h>
#include "csi_f.h"
#include "rif_f.h"
#include "cell.h"
#include "sif.h"

#define ULL	unsigned long long
#define LEN	2048
#define EPS	1e-9

static int	BPS,B_S=0;
static Cell	C;
static double	F1,F2,B,S;
static ULL	HG[1ULL<<(sizeof(Cell)*8)];

static void	T2T_Float(char *O,char *name,char *N)
{
	int	bps,n,X,Y,y,x;
	FOM	**fom;
	char	*id,str[8][LEN],*u_c,ID[LEN],
		*u="using integer PV as is floating-point PV",
		*c="converting PV by TIFF image description tag";
	double	f1,f2,f,b,s;
	Cell	**cell;

	if ((bps=ReadImageFile_Float(O,&X,&Y,&fom,&id))!=0) {
	    if (id!=NULL) {
		if ((n=sscanf(id,"%s %s %s %s %s %s %s %s",
				 str[0],str[1],str[2],str[3],
				 str[4],str[5],str[6],str[7]))==2 || n==6)
		    if (sscanf(str[n-2],"%lf",&f1)!=1 ||
			sscanf(str[n-1],"%lf",&f2)!=1 || f1>=f2)
			u_c=u;
		    else {
			u_c=c; b=f1; s=(f2-f1)/(double)((1ULL<<bps)-1);
		    }
		else
		    u_c=((n!=3 && n!=4 && n!=8) ||
			 sscanf(str[n-2],"%lf",&b)!=1 ||
			 sscanf(str[n-1],"%lf",&s)!=1 || s<=0.0)?u:c;

		if (u_c==c)
                    for (y=0; y<Y; y++)
		    for (x=0; x<X; x++) fom[y][x]=b+s*fom[y][x];

		free(id);
	    }
#ifdef	U_C
	    else
		u_c=u;

	    (void)fprintf(stderr,"%s : %s (warning).\n",O,u_c);
#endif
	}
	else
	    if (id!=NULL) free(id);

	if (!B_S) {
	    F1=F2=fom[0][0];
	    for (y=0; y<Y; y++)
	    for (x=0; x<X; x++)
		if ((f=fom[y][x])<F1) F1=f; else if (f>F2) F2=f;

	    B=F1;
	    if ((S=(F2-F1)/(double)C)<EPS) {
		(void)fputs("bad value step (warning).\n",stderr); S=1.0/EPS;
	    }
	}
	if (name!=NULL) (void)printf("%s\t%lf\t%le\n",name,B,S);

	cell=(Cell **)fom;
	for (y=0; y<Y; y++)
	for (x=0; x<X; x++)
	    ++HG[cell[y][x]=((f=fom[y][x])<F1)?0:(f>F2)?C:(Cell)((f-B)/S+0.5)];

	if (N!=NULL) {
	    (void)sprintf(ID,"%lf\t%lf",F1,F2);
	    StoreImageFile(N,X,Y,BPS,cell,ID);
	}
	while (--Y>=0) free(fom[Y]); free(fom);
}

#define NFD	argv[argc-1]

int	main(int argc,char **argv)
{
	ULL	i;
	char	**O,N[LEN];
	int	Z,z,len;

	if (argc<4 || argc>7) {
	(void)fputs("usage : t2t_float org.tif {base step} BPS new.tif\n"
		    "        t2t_float org/ nameFile {base step} BPS new/\n",
		    stderr); exit(1);
	}
	if ((BPS=atoi(argv[argc-2]))<=0 || BPS>sizeof(Cell)*8) {
	    (void)fputs("bad bits per sample (BPS).\n",stderr); exit(1);
	}
	C=(Cell)((1ULL<<BPS)-1);

	if (argc==6 || argc==7) {
	    B=atof(argv[argc-4]);
	    if ((S=atof(argv[argc-3]))<=0.0) {
		(void)fputs("bad value step.\n",stderr); exit(1);
	    }
	    F1=B; F2=B+S*(double)C; B_S=1;

	    if (NFD[0]=='-') for (i=0; i<=C; i++) HG[i]=0;
	}
	else
	    if (NFD[0]=='-') {
		(void)fputs("missing value base and step.\n",stderr); exit(1);
	    }

	if (argc==4 || argc==6)
	    if (NFD[0]=='-')
		if (NFD[1]=='\0')
		    T2T_Float(argv[1],NULL,NULL);
		else
		    T2T_Float(argv[1],NULL,NFD+1);
	    else
		T2T_Float(argv[1],NFD,NFD);
	else {
	    O=CheckSliceImages_Float(argv[1],argv[2],NULL,NULL,&Z,NULL);

	    if (NFD[0]!='-' || NFD[1]!='\0')
		for (len=0; argv[1][len]!='\0'; len++) ;

	    for (z=0; z<Z; z++) {
		if (NFD[0]=='-')
		    if (NFD[1]=='\0')
			T2T_Float(O[z],NULL,NULL);
		    else {
			(void)sprintf(N,"%s/%s",NFD+1,O[z]+len+1);
			T2T_Float(O[z],NULL,N);
		    }
		else {
		    (void)sprintf(N,"%s/%s",NFD,O[z]+len+1);
		    T2T_Float(O[z],O[z]+len+1,N);
		}
		free(O[z]);
	    }
	    free(O);
	}
	if (NFD[0]=='-')
	    for (i=0; i<=C; i++)
		if (HG[i]!=0)
		    (void)printf("%llu\t%lf\t%lf\t%llu\n",
				 i,B+S*((double)i-0.5),
				   B+S*((double)i+0.5),HG[i]);

	return 0;
}
