
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "rhp.h"
#include "cbp.h"
#include "sif_f.h"

extern void	Error(char *msg);

#ifdef	_WIN32
#include <process.h>
#include <windows.h>

#define THREAD_T	HANDLE
#define FUNCTION_T	unsigned __stdcall
#define RETURN_VALUE	0

#define INIT_THREAD(T,F,A)	\
	(T=(HANDLE)_beginthreadex(NULL,0,F,(void *)(A),0,NULL))==0
#define TERM_THREAD(T)	\
	WaitForSingleObject(T,INFINITE)==WAIT_FAILED || CloseHandle(T)==0
#else
#include <pthread.h>

#define THREAD_T	pthread_t
#define FUNCTION_T	void *
#define RETURN_VALUE	NULL

#define INIT_THREAD(T,F,A)	pthread_create(&(T),NULL,F,(void *)(A))
#define TERM_THREAD(T)		pthread_join(T,NULL)
#endif

#define INIT_MT(T,F,A)	\
if (INIT_THREAD(T,F,A)) Error("multi-threading initialization error.")
#define TERM_MT(T)	\
if (TERM_THREAD(T)) Error("multi-threading termination error.")

#define LEN	2048

typedef struct {
		char	form[LEN];
		int	z,N;
		FOM	**F;
	} Struct;

static FUNCTION_T	Store(void *a)
{
	Struct	*s=(Struct *)a;
	char	path[LEN];

	(void)sprintf(path,s->form,s->z);
	StoreImageFile_Float(path,s->N,s->N,s->F,SIF_F_desc);
	(void)printf("%d\t%s\n",s->z,SIF_F_desc);

	return RETURN_VALUE;
}

#define EPS	1e-9
#define DEG	3.14159265358979323846/180.0

int	main(int argc,char **argv)
{
	HiPic		hp;
	double		Dr,RC,RA0,y0,dy,Dz=0.0;
	int		Nz,z1,z2,y1,y2,y,z,t,x,len;
	Float		**P,***W,*w,*P_F,**F;
	Struct		S;
	FOM		*hp_T,fom,*S_F;
	THREAD_T	T;

	if (argc<6 || argc>9)
	    Error("usage : hp_tg HiPic/ Dr RC {Dz} {layer1 layer2} RA0 TG/");

	InitReadHiPic(argv[1],&hp);

	if (hp.Nt<2) Error("too few views.");

	if ((Dr=atof(argv[2]))<EPS)
	    Error("bad horizontal interval of detectors on HiPic image.");

	if ((RC=atof(argv[3]))<=0.0 || RC>=(double)(hp.Nx-1))
	    Error("bad rotation center (used as origin of detector array).");

	if (argc==6 || argc==8)
	    Nz=hp.Ny;
	else
	    if ((Nz=hp.Ny+(int)ceil(fabs(Dz=atof(argv[4]))))>=2*hp.Ny)
		Error("too large vertical displacement.");

	if (argc==6 || argc==7) {
	    z1=0; z2=Nz-1;
	}
	else
	    if ((z1=atoi(argv[argc-4]))<0 ||
		(z2=atoi(argv[argc-3]))>=Nz || z1>z2) Error("bad slice range.");

	RA0=atof(argv[argc-2])*DEG;

	P=InitCBP(hp.Nx,hp.Nt);

	if (Dz<0.0) {
	    y0=0.5+Dz; if ((y1=z1+(int)floor(y0))<0) y1=0;
	}
	else {
	    y0=0.5; if ((y1=z1+(int)floor(y0-Dz))<0) y1=0;
	}
	y=(y2=(z2<hp.Ny)?z2:hp.Ny-1)-y1+1;
	if ((W=(Float ***)malloc(sizeof(Float **)*y))==NULL ||
	    (*W=(Float **)malloc(sizeof(Float *)*y*hp.Nt))==NULL ||
	    (**W=(Float *)malloc(sizeof(Float)*y*hp.Nt*hp.Nx))==NULL ||
	    (S.F=(FOM **)malloc(sizeof(FOM *)*hp.Nx))==NULL ||
	    (*(S.F)=(FOM *)malloc(sizeof(FOM)*hp.Nx*hp.Nx))==NULL)
	    Error("memory allocation error.");

	for (z=0; z<y; z++) {
	    if (z!=0) W[z]=W[z-1]+hp.Nt;

	    for (t=z==0; t<hp.Nt; t++) W[z][t]=W[z][t-1]+hp.Nx;
	}
	for (y=1; y<hp.Nx; y++) S.F[y]=S.F[y-1]+hp.Nx;

	for (t=0; t<hp.Nt; t++) {
	    ReadHiPic(&hp,t);

	    for (y=y1; y<=y2; y++) {
		w=W[y-y1][t]; hp_T=hp.T[y];
		for (x=0; x<hp.Nx; x++)
		    *(w++)=((fom=(*(hp_T++)))>0.0)?-log(fom):0.0;
	    }
	}
	TermReadHiPic(&hp);

	S.N=hp.Nx;

	len=1; for (z=10; z<Nz; z*=10) ++len;
	(void)sprintf(S.form,"%s/%%0%dd.tif",argv[argc-1],len);

	dy=Dz/(double)hp.Nt; 
	for (z=z1; z<=z2; z++) {
	    for (t=0; t<hp.Nt; t++) {
		P_F=P[t];
		if ((y=z+(int)floor(y0-dy*(double)t))<y1 || y>y2)
		    for (x=0; x<hp.Nx; x++) *(P_F++)=0.0;
		else {
		    w=W[y-y1][t]; for (x=0; x<hp.Nx; x++) *(P_F++)=(*(w++));
		}
	    }
	    F=CBP(Dr,-RC,RA0);

	    if (z!=z1) TERM_MT(T);

	    S.z=z;

	    for (y=0; y<hp.Nx; y++) {
		S_F=S.F[y]; P_F=F[y];
		for (x=0; x<hp.Nx; x++) *(S_F++)=(*(P_F++));
	    }
	    INIT_MT(T,Store,&S);
	}
	TERM_MT(T);

	free(*(S.F)); free(S.F); free(**W); free(*W); free(W); TermCBP();

	return 0;
}
